/*
    Insanity Pong
    (C) 2006 Jason Hunt
    nulluser@gmail.com
*/


#include <windows.h>
#include <stdio.h>

#define UPDATE_TIMER            1           // ID for frame timer
#define FRAME_TIME              10
#define MAX_BALLS               100
#define MAX_BALL_SPEED          0.008       // Max ball speed
#define STAR_RANGE              -4
#define STAT_BAR_SIZE           20
#define MENU_SIZE               20          // Size of upper window bar
#define MAX_STARS               50



char program_name[] = "InsanityPong";

/* Position size and movment data for screen objects */
typedef struct
{
   double x, y, dx, dy;
   double x_size, y_size;
    
} motion_type;


/* Data for a star */
typedef struct
{
    double x, y, z;
    int old_sx, old_sy, sx, sy;
} star_type;

motion_type old_left_paddle, old_right_paddle, old_balls[MAX_BALLS];        // Old ball position
motion_type left_paddle, right_paddle, balls[MAX_BALLS];                    // Current ball position

int num_balls = 2;

star_type stars[MAX_STARS];      // Star data
int num_stars = MAX_STARS;       // Number of stars

int display_x = 640;             // Initial screen size
int display_y = 480;

bool clear_screen = true;        // True of full redraw is needed

int player_score = 0;            // Inital player scrore
int computer_score = 0;          // Inital computer score

HDC memdc = NULL;	           	 // Memory device context
HBITMAP h_mem_bitmap = NULL;     // Offscreen bitmap


/*********************
*** Init functions ***
*********************/


/* Reset all of the stars */
void reset_stars( void )
{
    for (int i = 0; i < num_stars; i++)
    {
        stars[i].x = (rand() % 2000) / 1000.0;
        stars[i].y = (rand() % 2000) / 1000.0;
        
        if (rand() % 2) stars[i].x *= -1;
        if (rand() % 2) stars[i].y *= -1;
        
        stars[i].z = STAR_RANGE * (rand() % 1000) / 1000.0;
    }
}
/* End of reset stars */


/* Rest a ball the center of the screen */
void reset_ball(motion_type &m)
{
    m.x = 0.5;                // range from [0, 1];
    m.y = 0.5;
    m.x_size = 0.03;
    m.y_size = 0.04;
    
    m.dx = (rand() % 100) / 10000.0 + 0.001;
    m.dy = (rand() % 100) / 10000.0 + 0.001;

    if (rand() % 2) m.dx *= -1;
    if (rand() % 2) m.dy *= -1;    
}
/* End of reset ball */


/* Adds a ball to the system */
void add_ball( void )
{
    if (num_balls >= MAX_BALLS) return;    
    
    reset_ball(balls[num_balls]);
    num_balls++;
}
/* End of add_ball */


/* Reset all game parameters */
void reset_motion( void )
{
    clear_screen = true;

    left_paddle.x_size = 0.008;
    left_paddle.y_size = 0.15;

    left_paddle.x = left_paddle.x_size * 2; 
    left_paddle.y = 0.5;    

    right_paddle.x_size = 0.008;
    right_paddle.y_size = 0.15;

    right_paddle.x = 1 - right_paddle.x_size * 2; 
    right_paddle.y = 0.5;    
        
    for (int i = 0; i < num_balls; i++)
        reset_ball(balls[i]);
}
/* End of reset game */


/****************************
*** End of Init functions ***
****************************/



/*************************
*** Graphics Functions ***
*************************/


/* Return a bounding box in screen coords */
void get_screen_coords(motion_type m, RECT &r)
{
    r.right = (long)((m.x - m.x_size / 2.0) * display_x);
    r.top = (long)((m.y - m.y_size / 2.0) * (display_y-STAT_BAR_SIZE));

    r.left = r.right + (long)(m.x_size * display_x);
    r.bottom = r.top + (long)(m.y_size * (display_y-STAT_BAR_SIZE)); 
}
/* End of get_screen_coords */


/* Draw game display */
void draw_game( void )
{
 
    RECT r;
    
    // Erase in black
    SelectObject(memdc, (HBRUSH) GetStockObject(BLACK_BRUSH));
    SelectObject(memdc, (HBRUSH) GetStockObject(BLACK_PEN));    
   
    if (clear_screen)
    {
        Rectangle(memdc, 0, 0, display_x, display_y);    
        clear_screen = false;       
    }
    
    // Erase old stars
    for (int s = 0; s < num_stars; s++)
    {
        int sx = stars[s].old_sx;
        int sy = stars[s].old_sy;
            
        if (sx > 0 && sx < display_x && sy > 0 && sy < display_y)
            Ellipse(memdc,sx, sy, sx+3, sy+3);
    }

    // Erase old balls 
    for (int i = 0; i < num_balls; i++)
    {
        get_screen_coords(old_balls[i], r);
        Ellipse(memdc, r.right, r.top, r.left, r.bottom);
    }    
    
    // Erase old left paddle
    get_screen_coords(old_left_paddle, r);    
    FillRect(memdc, &r, (HBRUSH) GetStockObject(BLACK_BRUSH));  
    
    // Erase old right paddle
    get_screen_coords(old_right_paddle, r);    
    FillRect(memdc, &r, (HBRUSH) GetStockObject(BLACK_BRUSH));  
        
    // Draw in white
    SelectObject(memdc, (HBRUSH) GetStockObject(WHITE_PEN));
    SelectObject(memdc, (HBRUSH) GetStockObject(WHITE_BRUSH));
    
    // Draw stars
    for (int s = 0; s < num_stars; s++)
    {
        int sx = stars[s].sx;
        int sy = stars[s].sy;
        
        if (sx > 0 && sx < display_x && sy > 0 && sy < display_y)
            Ellipse(memdc,sx, sy, sx+3, sy+3);
    }

    // Draw balls    
    for (int i = 0; i < num_balls; i++)
    {
        get_screen_coords(balls[i], r);            
        Ellipse(memdc, r.right, r.top, r.left, r.bottom);
    }

    // Draw left paddle
    get_screen_coords(left_paddle, r);    
    FillRect(memdc, &r, (HBRUSH) GetStockObject(WHITE_BRUSH));  
    
    // Draw right paddle
    get_screen_coords(right_paddle, r);    
    FillRect(memdc, &r, (HBRUSH) GetStockObject(WHITE_BRUSH));  
    
    

    // Status text
    char b[400];    
    sprintf(b, "Comp: %d  ", computer_score);
    TextOut(memdc, 1, display_y - STAT_BAR_SIZE, b, strlen(b));
    
    sprintf(b, "Player: %d  ", player_score);
    TextOut(memdc, display_x - 75, display_y - STAT_BAR_SIZE, b, strlen(b));

     
}
/* End of draw_game */




/* Create off screen buffer */
void setup_display( HWND hwnd )
{
    // Get main DC
    HDC hdc = GetDC(hwnd);

  	// Create Memory device contextand bitmap
    memdc = CreateCompatibleDC(hdc);	

    // Create off screen buffer
  	h_mem_bitmap = CreateCompatibleBitmap(hdc, display_x, display_y);

  	// Select the off screen buffer into the DC
   	SelectObject(memdc, h_mem_bitmap);		

    ReleaseDC(hwnd, hdc);       // Main DC is no longer needed
}
/* End of setup_display */



/* Resize the display and update buffers */
void resize_display(HWND hwnd, LPARAM l)
{
    display_x = LOWORD(l);
    display_y = HIWORD(l);

    DeleteObject(h_mem_bitmap);		//	 Delete old screen buffer
    DeleteDC(memdc);		      	// Delete old memory DC

    setup_display(hwnd);

    InvalidateRect(hwnd, NULL, true);   
}
/* End of resize_display */


/********************************
*** End of Graphics Functions ***
********************************/



/*************************
*** Movement functions ***
*************************/


/* Update player paddle with mouse coords */
void update_paddles(HWND hwnd, WPARAM w, LPARAM l)
{
    // Save old position for erase    
    old_right_paddle = right_paddle;

    right_paddle.y = HIWORD(l) / (double)display_y;
    
    right_paddle.x =( 3*display_x / 4 + LOWORD(l)/4) / ((double)display_x);
        
    if (right_paddle.y < right_paddle.y_size / 2) 
        right_paddle.y = right_paddle.y_size / 2;
        
    if (right_paddle.y > 1 - right_paddle.y_size / 2) 
        right_paddle.y = 1 - right_paddle.y_size / 2;

    draw_game();
}
/* End of update paddles */


/* Move all screen opjects */
void move_objects( HWND hwnd )
{
    // Save old left paddle now because it gets updated
    // based on the position of each ball
    old_left_paddle = left_paddle;

    // Save old ball position for erase
    for (int i = 0; i < num_balls; i++)   
    {
        old_balls[i] = balls[i];   
    
        // Create a small random vector
        double ball_mod_x = (rand() % 10000) / 4000000.0;
        double ball_mod_y = (rand() % 10000) / 4000000.0;

        // Randomize direction
        if (rand() % 2) ball_mod_x *= -1;
        if (rand() % 2) ball_mod_y *= -1;
    
        // Add insanity factor
        balls[i].dx += ball_mod_x;
        balls[i].dy += ball_mod_y;
    
        // Clip balls[i] speed
        if (balls[i].dx > MAX_BALL_SPEED ) balls[i].dx = MAX_BALL_SPEED ;
        if (balls[i].dx < -MAX_BALL_SPEED ) balls[i].dx = -MAX_BALL_SPEED ;

        if (balls[i].dy > MAX_BALL_SPEED ) balls[i].dy = MAX_BALL_SPEED;
        if (balls[i].dy < -MAX_BALL_SPEED ) balls[i].dy = -MAX_BALL_SPEED;

        // Move balls[i]
        balls[i].x += balls[i].dx;
        balls[i].y += balls[i].dy;
        
        // Check collision with right paddle
        if ((balls[i].x + balls[i].x_size / 2 > right_paddle.x - right_paddle.x_size / 2) &&
            (balls[i].y > right_paddle.y - right_paddle.y_size / 2) &&
            (balls[i].y < right_paddle.y + right_paddle.y_size / 2) &&
            balls[i].x < right_paddle.x)
            {
                balls[i].dx *= -1;
                balls[i].x -= left_paddle.x_size;   
            }
                
       // Check collision with left paddle
        if ((balls[i].x - balls[i].x_size / 2 < left_paddle.x + left_paddle.x_size / 2) &&
            (balls[i].y > left_paddle.y - left_paddle.y_size / 2) &&
            (balls[i].y < left_paddle.y + left_paddle.y_size / 2) &&
             balls[i].x > left_paddle.x)
            {
                balls[i].dx *= -1;
                balls[i].x += left_paddle.x_size;   
            }
    
        // Right player score
        if (balls[i].x - balls[i].x_size / 2 < 0) 
        { 
            player_score++;
            reset_ball(balls[i]);
            return;
        }

        // Left player score
        if (balls[i].x + balls[i].x_size / 2 > 1) 
        {
            computer_score++;
            reset_ball(balls[i]);
            return;
        }
    
        // Reflect balls[i] off top
        if (balls[i].y - balls[i].x_size / 2 < 0) 
        { 
            balls[i].y = balls[i].x_size / 2; 
            balls[i].dy *= -1;
        }

        // Reflect balls[i] off bottom
        if (balls[i].y + balls[i].y_size / 2 > 1) 
        { 
            balls[i].y = 1 - balls[i].y_size / 2; 
            balls[i].dy *= -1;
        }    

        // Move computer controlled paddle
        
        // Seek to balls[i] y
        if (balls[i].y > left_paddle.y)
            left_paddle.y += 0.004;
    
        if (balls[i].y < left_paddle.y)
            left_paddle.y -= 0.004;
    
        // Range check
        if (left_paddle.y - left_paddle.y_size / 2 < 0)
            left_paddle.y =left_paddle.y_size / 2;
    
        if (left_paddle.y + left_paddle.y_size / 2 > 1)
            left_paddle.y = 1 - left_paddle.y_size / 2;
    }   

    // Force redraw 
    InvalidateRect(hwnd, NULL, 0);
}
/* End of move objects */


/* Move all of the stars and calculate screen position */
void move_stars ( void )
{
    // Move stars
    for (int i = 0; i < num_stars; i++)
    {
        stars[i].z -= 0.04;
        
        // Make sure the srat is still in the display range
        if (stars[i].z < STAR_RANGE) 
        {
            // Create new position
            stars[i].x = (rand() % 2000) / 1000.0;
            stars[i].y = (rand() % 2000) / 1000.0;
            
            // Randomize over directions
            if (rand() % 2) stars[i].x *= -1;
            if (rand() % 2) stars[i].y *= -1;            
            
            // Reset z position
            stars[i].z  = -1;
        }   

        // The star has moved, save old screen position
        stars[i].old_sx = stars[i].sx;
        stars[i].old_sy = stars[i].sy;

        // Calculate new screen position
        stars[i].sx = (int)((256 * stars[i].x / stars[i].z) + display_x / 2);
        stars[i].sy = (int)(display_y / 2 - (256 * stars[i].y / stars[i].z));            
    }    
}
/* End of move stars */



/********************************
*** End of Movement functions ***
********************************/



/* The window Function */
LRESULT CALLBACK WinProc(HWND hwnd, UINT msg, WPARAM w, LPARAM l)
{
    HDC hdc;

    switch(msg)
    {
        case WM_CREATE: 
            reset_stars();
            reset_motion(); 

            setup_display(hwnd);
            
            ShowCursor(false);

            SetTimer(hwnd, UPDATE_TIMER, FRAME_TIME, NULL);
            return(0);

        case WM_CHAR:
            if (w == 27) DestroyWindow(hwnd);   // Escape exits
            if (w == VK_SPACE) add_ball();      // Space adds balls
            return(0);

        case WM_PAINT:
            PAINTSTRUCT ps;
            hdc = BeginPaint(hwnd, &ps);
            BitBlt(hdc, 0, 0, display_x, display_y, memdc, 0, 0, SRCCOPY);
            EndPaint(hwnd, &ps);  
           return(0);

        case WM_TIMER:
            move_objects(hwnd);
            move_stars();
            draw_game();

//            InvalidateRect(hwnd, NULL, 0);
            return(0);

        case WM_DESTROY:
            DeleteObject(h_mem_bitmap);		//	 Delete off screen buffer
            DeleteDC(memdc);	       		// Delete memory DC
    
            ShowCursor(true);
            PostQuitMessage(0);
            return(0);
            
        case WM_MOUSEMOVE:
            
            update_paddles(hwnd, w, l);
            return(0);
            
        case WM_SIZE:
            resize_display(hwnd, l);
            return(0);

        default:
            return(DefWindowProc(hwnd, msg, w, l));
    }

    return(0);  
}
/* End of WinProc */


/* WinMain */
int WINAPI WinMain(HINSTANCE h_this_inst, HINSTANCE h_prev_inst, LPSTR args, int mode)
{
    WNDCLASS wnd_class;
    
    wnd_class.hInstance = h_this_inst;
    wnd_class.lpszClassName = program_name;
    wnd_class.lpfnWndProc = WinProc;
    wnd_class.style = 0;
    
    wnd_class.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wnd_class.hCursor = LoadCursor(NULL, IDC_ARROW);
    wnd_class.lpszMenuName = NULL;
    
    wnd_class.cbClsExtra = 0;
    wnd_class.cbWndExtra = 0;
    
    wnd_class.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    
    if (!RegisterClass(&wnd_class)) return(0);
        
    // Create the window
    HWND hwnd = CreateWindow(program_name, "Insanity Pong (c) 2006 Jason Hunt", WS_OVERLAPPEDWINDOW,
                             300, 200,   // Pos
                             display_x, display_y + MENU_SIZE,   // Size
                             HWND_DESKTOP, NULL, h_this_inst, NULL);

    ShowWindow(hwnd, mode);
    UpdateWindow(hwnd);
           
    MSG msg;
    
    // Windows message loop
    while(GetMessage(&msg, NULL, 0, 0))
    { 
        TranslateMessage(&msg);
        DispatchMessage(&msg);                            
    }
                        
    return(msg.wParam);
}
/* End of winmain */



